{%- extends "basic/layout.html" %}

{%- block sidebarlogo %}
{%- if logo %}
  <div class="logo">
    <a href="{{ pathto(master_doc) }}">
      <img class="logo" src="{{ pathto('_static/' + logo, 1) }}" alt="Logo"/>
    </a>
  </div>
{%- endif %}
{%- endblock %}

